import datetime

from django.core.exceptions import ValidationError
from django.utils import timezone
from django.test import TestCase

from pinax.types.values import VALUE_TYPES
from pinax.types.periods import PERIOD_TYPES, validate, parse, get_period, period_for_date, period_range, period_display, period_start_end


class ValueTypesTests(TestCase):

    def test_integer_value_type_raises_error(self):
        with self.assertRaises(ValidationError):
            VALUE_TYPES["integer"].validate("foo")

    def test_integer_value_type_validates(self):
        self.assertIsNone(VALUE_TYPES["integer"].validate("566"))

    def test_boolean_value_type_raises_error_on_not_true_or_false(self):
        with self.assertRaises(ValidationError):
            VALUE_TYPES["boolean"].validate("foo")

    def test_boolean_value_type_validates_true(self):
        self.assertIsNone(VALUE_TYPES["boolean"].validate("true"))

    def test_boolean_value_type_validates_false(self):
        self.assertIsNone(VALUE_TYPES["boolean"].validate("false"))

    def test_decimal_value_type_raises_error(self):
        with self.assertRaises(ValidationError):
            VALUE_TYPES["decimal"].validate("foo")

    def test_decimal_value_type_validates(self):
        self.assertIsNone(VALUE_TYPES["decimal"].validate("5.66"))

    def test_monetary_value_type_raises_error(self):
        with self.assertRaises(ValidationError):
            VALUE_TYPES["monetary"].validate("foo")

    def test_monetary_value_type_validates(self):
        self.assertIsNone(VALUE_TYPES["monetary"].validate("56.60"))

    def test_hour_value_type_raises_error(self):
        with self.assertRaises(ValidationError):
            VALUE_TYPES["hours"].validate("foo")

    def test_hour_value_type_validates(self):
        self.assertIsNone(VALUE_TYPES["hours"].validate("56"))

    def test_traffic_light_value_type_raises_error_on_big_int(self):
        with self.assertRaises(ValidationError):
            VALUE_TYPES["traffic-light"].validate("1000")

    def test_traffic_light_value_type_raises_error_on_non_int(self):
        with self.assertRaises(ValidationError):
            VALUE_TYPES["traffic-light"].validate("foo")

    def test_traffic_light_value_type_validates(self):
        self.assertIsNone(VALUE_TYPES["traffic-light"].validate("2"))

    def test_monetary_value_display(self):
        self.assertEquals(VALUE_TYPES["monetary"].display(1000), "$1,000")

    def test_percentage_value_display(self):
        self.assertEquals(VALUE_TYPES["percentage"].display(0.37), "37.0%")

    def test_hours_value_display(self):
        self.assertEquals(VALUE_TYPES["hours"].display(1000), "1000h")

    def test_traffic_light_value_display(self):
        self.assertEquals(VALUE_TYPES["traffic-light"].display(1), "red")


class PeriodTests(TestCase):

    def setUp(self):
        self.quarter_1 = get_period("Q-2015-1")
        self.quarter_2 = get_period("Q-2015-2")
        self.year = get_period("Y-2015")

    def test_quarterly_sub_periods_monthly_all(self):
        periods = self.quarter_1.sub_periods("monthly")
        self.assertEquals(len(periods), 3)
        self.assertEquals(periods[0].get_display(), "January 2015")
        self.assertEquals(periods[1].get_display(), "February 2015")
        self.assertEquals(periods[2].get_display(), "March 2015")

    def test_quarterly_sub_periods_weekly_all(self):
        periods = self.quarter_1.sub_periods("weekly")
        self.assertEquals(len(periods), 14)
        self.assertEquals(periods[0].raw_value, "W-2015-01")
        self.assertEquals(periods[1].raw_value, "W-2015-02")
        self.assertEquals(periods[2].raw_value, "W-2015-03")
        self.assertEquals(periods[3].raw_value, "W-2015-04")
        self.assertEquals(periods[4].raw_value, "W-2015-05")
        self.assertEquals(periods[5].raw_value, "W-2015-06")
        self.assertEquals(periods[6].raw_value, "W-2015-07")
        self.assertEquals(periods[7].raw_value, "W-2015-08")
        self.assertEquals(periods[8].raw_value, "W-2015-09")
        self.assertEquals(periods[9].raw_value, "W-2015-10")
        self.assertEquals(periods[10].raw_value, "W-2015-11")
        self.assertEquals(periods[11].raw_value, "W-2015-12")
        self.assertEquals(periods[12].raw_value, "W-2015-13")
        self.assertEquals(periods[13].raw_value, "W-2015-14")

    def test_yearly_sub_periods_quarterly_all(self):
        periods = get_period("Y-2015").sub_periods("quarterly")
        self.assertEquals(len(periods), 4)
        self.assertEquals(periods[0].raw_value, "Q-2015-1")
        self.assertEquals(periods[1].raw_value, "Q-2015-2")
        self.assertEquals(periods[2].raw_value, "Q-2015-3")
        self.assertEquals(periods[3].raw_value, "Q-2015-4")

    def test_yearly_sub_periods_monthly_all(self):
        periods = get_period("Y-2015").sub_periods("monthly")
        self.assertEquals(len(periods), 12)
        self.assertEquals(periods[0].get_display(), "January 2015")
        self.assertEquals(periods[1].get_display(), "February 2015")
        self.assertEquals(periods[2].get_display(), "March 2015")
        self.assertEquals(periods[3].get_display(), "April 2015")
        self.assertEquals(periods[4].get_display(), "May 2015")
        self.assertEquals(periods[5].get_display(), "June 2015")
        self.assertEquals(periods[6].get_display(), "July 2015")
        self.assertEquals(periods[7].get_display(), "August 2015")
        self.assertEquals(periods[8].get_display(), "September 2015")
        self.assertEquals(periods[9].get_display(), "October 2015")
        self.assertEquals(periods[10].get_display(), "November 2015")
        self.assertEquals(periods[11].get_display(), "December 2015")

    def test_yearly_sub_periods_weekly_all(self):
        periods = get_period("Y-2015").sub_periods("weekly")
        self.assertEquals(len(periods), 53)
        for i in range(53):
            self.assertEquals(periods[i].raw_value, "W-2015-{:02d}".format(i + 1))

    def test_monthly_sub_periods_weekly_all(self):
        periods = get_period("M-2015-01").sub_periods("weekly")
        self.assertEquals(len(periods), 5)
        for i in range(5):
            self.assertEquals(periods[i].raw_value, "W-2015-{:02d}".format(i + 1))

    def test_validate_for_pass(self):
        try:
            self.quarter_1.validate_for("quarterly")
        except ValidationError:
            self.fail()

    def test_validate_for_fail(self):
        with self.assertRaises(ValidationError):
            self.quarter_1.validate_for("weekly")

    def test_validate_can_contain_type_pass_1(self):
        try:
            self.quarter_1.validate_can_contain_type("quarterly")
        except ValidationError:
            self.fail()

    def test_validate_can_contain_type_pass_2(self):
        try:
            self.quarter_1.validate_can_contain_type("weekly")
        except ValidationError:
            self.fail()

    def test_validate_can_contain_type_faik(self):
        with self.assertRaises(ValidationError):
            self.quarter_1.validate_can_contain_type("yearly")

    def test_get_display(self):
        self.assertEquals(get_period("M-2015-04").get_display(), "April 2015")

    def test_current(self):
        period = get_period(PERIOD_TYPES["quarterly"].for_date(timezone.now()))
        self.assertTrue(period.is_current())

    def test_past(self):
        period = get_period("M-2013-11")
        self.assertTrue(period.is_past())

    def test_includes_1(self):
        self.assertTrue(self.year.includes(self.quarter_1))
        self.assertTrue(self.year.includes(self.quarter_2))

    def test_includes_2(self):
        self.assertFalse(self.quarter_1.includes(self.year))

    def test_includes_3(self):
        self.assertFalse(self.quarter_1.includes(self.quarter_2))

    def test_includes_4(self):
        self.assertTrue(self.quarter_1.includes(get_period("Q-2015-1")))

    def test_validate_random_string(self):
        with self.assertRaises(ValidationError):
            validate("Patrick")

    def test_equality_true(self):
        self.assertTrue(self.quarter_1 == get_period("Q-2015-1"))

    def test_equality_false_1(self):
        self.assertFalse(self.quarter_1 == self.quarter_2)

    def test_equality_false_2(self):
        self.assertFalse(self.quarter_1 == self.year)

    def test_non_equality_false(self):
        self.assertFalse(self.quarter_1 != get_period("Q-2015-1"))

    def test_non_equality_true_1(self):
        self.assertTrue(self.quarter_1 != self.quarter_2)

    def test_non_equality_true_2(self):
        self.assertTrue(self.quarter_1 != self.year)

    def test_less_than_true(self):
        self.assertTrue(self.quarter_1 < self.quarter_2)

    def test_less_than_false_1(self):
        self.assertFalse(self.quarter_2 < self.quarter_1)

    def test_less_than_false_2(self):
        self.assertFalse(self.quarter_1 < self.year)

    def test_greater_than_true(self):
        self.assertTrue(self.quarter_2 > self.quarter_1)

    def test_greater_than_false_1(self):
        self.assertFalse(self.quarter_1 > self.quarter_2)

    def test_greater_than_false_2(self):
        self.assertFalse(self.quarter_1 > self.year)

    def test_less_than_or_equal_true(self):
        self.assertTrue(self.quarter_1 <= self.quarter_2)

    def test_less_than_or_equal_false_1(self):
        self.assertFalse(self.quarter_2 <= self.quarter_1)

    def test_less_than_or_equal_false_2(self):
        self.assertFalse(self.quarter_1 <= self.year)

    def test_greater_than_or_equal_true(self):
        self.assertTrue(self.quarter_2 >= self.quarter_1)

    def test_greater_than_or_equal_false_1(self):
        self.assertFalse(self.quarter_1 >= self.quarter_2)

    def test_greater_than_or_equal_false_2(self):
        self.assertFalse(self.quarter_1 >= self.year)

    def test_parse_week_1(self):
        self.assertEquals(parse("2015-W03"), "W-2015-03")

    def test_parse_week_2(self):
        self.assertEquals(parse("2015-W3"), "W-2015-03")

    def test_parse_week_3(self):
        self.assertEquals(parse("2015W03"), "W-2015-03")

    def test_parse_week_4(self):
        self.assertEquals(parse("2015W3"), "W-2015-03")

    def test_parse_month_1(self):
        self.assertEquals(parse("jan 2015"), "M-2015-01")

    def test_parse_month_2(self):
        self.assertEquals(parse("Jan 2015"), "M-2015-01")

    def test_parse_month_3(self):
        self.assertEquals(parse("January 2015"), "M-2015-01")

    def test_parse_month_4(self):
        self.assertEquals(parse("1/2015"), "M-2015-01")

    def test_parse_month_5(self):
        self.assertEquals(parse("01/2015"), "M-2015-01")

    def test_parse_month_6(self):
        self.assertEquals(parse("2015 January"), "M-2015-01")

    def test_parse_quarter_1(self):
        self.assertEquals(parse("2015Q1"), "Q-2015-1")

    def test_parse_quarter_2(self):
        self.assertEquals(parse("2015Q1"), "Q-2015-1")

    def test_parse_quarter_3(self):
        with self.assertRaises(ValidationError):
            self.assertEquals(parse("2015Q5"), "Q-2015-5")

    def test_parse_year_1(self):
        self.assertEquals(parse("2015"), "Y-2015")

    def test_parse_invalid(self):
        self.assertIsNone(parse("Patrick"))

    def test_get_period_str(self):
        period = get_period("M-2015-01")
        self.assertEquals(str(period), "M-2015-01")

    def test_get_period_validation(self):
        with self.assertRaises(ValidationError):
            get_period("2013W22")

    def test_weekly_period_type_raises_error_wrong_format(self):
        with self.assertRaises(ValidationError):
            PERIOD_TYPES["weekly"].validate("2013W22")

    def test_weekly_period_type_raises_error_week_number_too_low(self):
        with self.assertRaises(ValidationError):
            PERIOD_TYPES["weekly"].validate("W-2013-00")

    def test_weekly_period_type_raises_error_week_number_too_high(self):
        with self.assertRaises(ValidationError):
            PERIOD_TYPES["weekly"].validate("W-2013-75")

    def test_weekly_period_type_validates_week(self):
        self.assertIsNone(PERIOD_TYPES["weekly"].validate("W-2013-22"))

    def test_monthly_period_type_raises_error_wrong_format(self):
        with self.assertRaises(ValidationError):
            PERIOD_TYPES["monthly"].validate("201312")

    def test_monthly_period_type_raises_error_wrong_week_number(self):
        with self.assertRaises(ValidationError):
            PERIOD_TYPES["monthly"].validate("M-2013-15")

    def test_monthly_period_type_validates_week(self):
        self.assertIsNone(PERIOD_TYPES["monthly"].validate("M-2013-12"))

    def test_quarterly_period_type_raises_error_wrong_format(self):
        with self.assertRaises(ValidationError):
            PERIOD_TYPES["quarterly"].validate("2013Q4")

    def test_quarterly_period_type_raises_error_wrong_week_number(self):
        with self.assertRaises(ValidationError):
            PERIOD_TYPES["quarterly"].validate("Q-20139")

    def test_quarterly_period_type_validates_week(self):
        self.assertIsNone(PERIOD_TYPES["quarterly"].validate("Q-2013-3"))

    def test_yearly_period_type_raises_error_wrong_format(self):
        with self.assertRaises(ValidationError):
            PERIOD_TYPES["yearly"].validate("2013")

    def test_yearly_period_type_raises_error_wrong_week_number(self):
        with self.assertRaises(ValidationError):
            PERIOD_TYPES["yearly"].validate("Y-20139")

    def test_yearly_period_type_validates_week(self):
        self.assertIsNone(PERIOD_TYPES["yearly"].validate("Y-2013"))

    def test_weekly_period_for_date_end_of_year(self):
        self.assertEquals(period_for_date("weekly", datetime.date(2014, 12, 27)), "W-2014-52")

    def test_weekly_period_for_date_start_of_year(self):
        self.assertEquals(period_for_date("weekly", datetime.date(2014, 12, 30)), "W-2015-01")

    def test_weekly_period_for_date(self):
        self.assertEquals(period_for_date("weekly", datetime.date(2013, 8, 7)), "W-2013-32")

    def test_quarterly_period_for_date(self):
        self.assertEquals(period_for_date("quarterly", datetime.date(2013, 8, 7)), "Q-2013-3")

    def test_monthly_period_for_date(self):
        self.assertEquals(period_for_date("monthly", datetime.date(2013, 8, 7)), "M-2013-08")

    def test_yearly_period_for_date(self):
        self.assertEquals(period_for_date("yearly", datetime.date(2013, 8, 7)), "Y-2013")

    def test_period_for_date_today(self):
        self.assertEquals(period_for_date("yearly"), "Y-{}".format(datetime.date.today().year))

    def test_weekly_period_range(self):
        self.assertEquals(list(period_range("W-2012-50", "W-2013-03")), ["W-2012-50", "W-2012-51", "W-2012-52", "W-2013-01", "W-2013-02"])
        self.assertEquals(list(period_range("W-2014-50", "W-2015-03")), ["W-2014-50", "W-2014-51", "W-2014-52", "W-2015-01", "W-2015-02"])
        self.assertEquals(list(period_range("W-2015-50", "W-2016-03")), ["W-2015-50", "W-2015-51", "W-2015-52", "W-2015-53", "W-2016-01", "W-2016-02"])
        self.assertEquals(list(period_range("W-2016-50", "W-2017-03")), ["W-2016-50", "W-2016-51", "W-2016-52", "W-2017-01", "W-2017-02"])

    def test_weekly_period_inclusive_range(self):
        self.assertEquals(list(period_range("W-2012-50", "W-2013-03", inclusive=True)), ["W-2012-50", "W-2012-51", "W-2012-52", "W-2013-01", "W-2013-02", "W-2013-03"])
        self.assertEquals(list(period_range("W-2014-50", "W-2015-03", inclusive=True)), ["W-2014-50", "W-2014-51", "W-2014-52", "W-2015-01", "W-2015-02", "W-2015-03"])
        self.assertEquals(list(period_range("W-2015-50", "W-2016-03", inclusive=True)), ["W-2015-50", "W-2015-51", "W-2015-52", "W-2015-53", "W-2016-01", "W-2016-02", "W-2016-03"])
        self.assertEquals(list(period_range("W-2016-50", "W-2017-03", inclusive=True)), ["W-2016-50", "W-2016-51", "W-2016-52", "W-2017-01", "W-2017-02", "W-2017-03"])

    def test_monthly_period_range(self):
        self.assertEquals(list(period_range("M-2012-11", "M-2013-03")), ["M-2012-11", "M-2012-12", "M-2013-01", "M-2013-02"])

    def test_monthly_period_inclusive_range(self):
        self.assertEquals(list(period_range("M-2012-11", "M-2013-03", inclusive=True)), ["M-2012-11", "M-2012-12", "M-2013-01", "M-2013-02", "M-2013-03"])

    def test_quarterly_period_range(self):
        self.assertEquals(list(period_range("Q-2012-3", "Q-2013-2")), ["Q-2012-3", "Q-2012-4", "Q-2013-1"])

    def test_quarterly_period_inclusive_range(self):
        self.assertEquals(list(period_range("Q-2012-3", "Q-2013-2", inclusive=True)), ["Q-2012-3", "Q-2012-4", "Q-2013-1", "Q-2013-2"])

    def test_yearly_period_range(self):
        self.assertEquals(list(period_range("Y-2010", "Y-2013")), ["Y-2010", "Y-2011", "Y-2012"])

    def test_yearly_period_range_inclusive(self):
        self.assertEquals(list(period_range("Y-2010", "Y-2013", inclusive=True)), ["Y-2010", "Y-2011", "Y-2012", "Y-2013"])

    def test_period_range_mismatched_types(self):
        with self.assertRaises(ValidationError):
            period_range("W-2012-50", "Y-2013")

    def test_yearly_period_type_display(self):
        self.assertEquals(period_display("Y-2013"), "2013")

    def test_monthly_period_type_display(self):
        self.assertEquals(period_display("M-2013-08"), "August 2013")

    def test_quarterly_period_type_display(self):
        self.assertEquals(period_display("Q-2013-3"), "2013Q3")

    def test_weekly_period_type_display(self):
        self.assertEquals(period_display("W-2013-32"), "Week of Aug 05, 2013")

    def test_yearly_period_type_start_end(self):
        self.assertEquals(period_start_end("Y-2013"), (datetime.date(2013, 1, 1), datetime.date(2013, 12, 31)))

    def test_monthly_period_type_start_end(self):
        self.assertEquals(period_start_end("M-2013-08"), (datetime.date(2013, 8, 1), datetime.date(2013, 8, 31)))

    def test_quarterly_period_type_start_end(self):
        self.assertEquals(period_start_end("Q-2013-3"), (datetime.date(2013, 7, 1), datetime.date(2013, 9, 30)))

    def test_weekly_period_type_start_end(self):
        self.assertEquals(period_start_end("W-2013-32"), (datetime.date(2013, 8, 5), datetime.date(2013, 8, 11)))



import calendar
import datetime
import operator
import re

from django.core.exceptions import ValidationError

from dateutil.parser import parse as dateutil_parse


"""
Period Types define different periods over which metrics can apply, e.g. weeks,
months or quarters.

 * PeriodType.validate(period) validates whether a string representation is
   valid for this period type
 * PeriodType.for_date(date) converts a datetime.date to the string
   representation for the period included by that date
 * PeriodType.start_end(period) returns a tuple of start date and end date for
   the given period
 * PeriodType.range(start, end) yields the periods from start to (but not
   including) stop. For example, period_range("Y-2010", "Y-2013") will yield
   "Y-2010", "Y-2011", "Y-2012".
 * PeriodType.display(period) displays the given period in a human-readable
   format

PERIOD_TYPES in this module maps the labels used for Period Types into the
classes themselves.

There are helper functions which dispatch to the right PeriodType for a given
period and call a class method on them:

 * period_start_end(period)
 * period_range(start, stop)
 * period_display(period)

There is also a helper function period_for_date which takes a period type name
like "weekly" and returns the period of the given date (or today if no date
given).
"""


class Period(object):  # abstract base class

    prefix = None
    validation_regex = r".*"
    minimum = None
    maximum = None
    contains = []

    def __init__(self, raw_value):
        self.validate(raw_value)
        self.raw_value = raw_value

    def __str__(self):
        return self.raw_value

    def includes(self, period):
        """
        quarter.includes(month) : True|False
        """
        if self == period:
            return True
        if period.prefix in self.contains:
            start, end = self.start_end(self.raw_value)
            start2, end2 = period.start_end(period.raw_value)
            return start <= start2 and end >= end2
        return False

    def sub_periods(self, period_type):
        self.validate_can_contain_type(period_type)
        periods = []
        if self.is_period_type(period_type):
            periods.append(get_period(self.raw_value))
        klass = PERIOD_TYPES[period_type]
        start, end = self.get_start_end()
        start, end = klass.for_date(start), klass.for_date(end)
        for period_raw in klass.range(start, end, inclusive=True):
            periods.append(get_period(period_raw))
        return periods

    def validate_can_contain_type(self, period_type):
        valids = [x.upper() for x in (self.contains + [self.prefix])]
        if period_type[0].upper() not in valids:
            raise ValidationError(
                "Periods of the type {} cannot exist within {}".format(
                    period_type,
                    ", ".join(valids)
                )
            )

    def is_period_type(self, period_type):
        return period_type[0].upper() == self.prefix.upper()

    def validate_for(self, period_type):
        if not self.is_period_type(period_type):
            raise ValidationError(
                "{} is must match this period's ({}) type".format(
                    period_type,
                    self.get_display()
                )
            )

    def get_start_end(self):
        return self.start_end(self.raw_value)

    def get_display(self):
        return self.display(self.raw_value)

    def is_past(self):
        current = self.for_date(datetime.datetime.now())
        current_period = get_period(current)
        return current_period > self

    def is_current(self):
        current = self.for_date(datetime.datetime.now())
        current_period = get_period(current)
        return current_period == self

    @classmethod
    def validate(cls, period):
        regex = "^{}-".format(cls.prefix) + cls.validation_regex
        match = re.match(regex, period)
        if not match:
            raise ValidationError(
                "Incorrect value: {0}".format(period)
            )
        if match.groups():
            part = int(match.groups()[0])
            if cls.minimum and part < cls.minimum:
                raise ValidationError(
                    "Incorrect value: {0}".format(period)
                )
            if cls.maximum and part > cls.maximum:
                raise ValidationError(
                    "Incorrect value: {0}".format(period)
                )

    def __eq__(self, other):
        return type(self) == type(other) and self.raw_value == other.raw_value

    def __ne__(self, other):
        return type(self) != type(other) or self.raw_value != other.raw_value

    def __lt__(self, other):
        # @@@ we might want to be smart and be able to do things like "Q-2014-Q4" < "Y-2015"
        return type(self) == type(other) and self.raw_value < other.raw_value

    def __gt__(self, other):
        return type(self) == type(other) and self.raw_value > other.raw_value

    def __le__(self, other):
        return type(self) == type(other) and self.raw_value <= other.raw_value

    def __ge__(self, other):
        return type(self) == type(other) and self.raw_value >= other.raw_value


class WeeklyPeriod(Period):

    prefix = "W"
    validation_regex = r"\d{4}-(\d{2})$"
    minimum = 1
    maximum = 53

    @classmethod
    def from_parts(cls, year, week):
        return "{}-{:d}-{:02d}".format(cls.prefix, int(year), int(week))

    @classmethod
    def for_date(cls, date):
        year = date.isocalendar()[0]
        week = date.isocalendar()[1]
        return cls.from_parts(year, week)

    @classmethod
    def start_end(cls, period):
        year = int(period[2:6])
        week = int(period[7:])
        start = iso_week_to_gregorian(year, week)
        end = start + datetime.timedelta(days=6)
        return start, end

    @classmethod
    def range(cls, start, stop, inclusive=False):
        cls.validate(start)
        cls.validate(stop)
        year_start = int(start[2:6])
        week_start = int(start[7:])
        year_stop = int(stop[2:6])
        week_stop = int(stop[7:])
        year = year_start
        week = week_start
        op = operator.le if inclusive else operator.lt
        while op((year, week), (year_stop, week_stop)):
            yield "{}-{:d}-{:02d}".format(cls.prefix, year, week)
            week += 1
            if datetime.date(year, 12, 31).isocalendar()[1] == 53:
                weeks_in_year = 53
            else:
                weeks_in_year = 52
            if week == weeks_in_year + 1:
                week = 1
                year += 1
        raise StopIteration()

    @classmethod
    def display(cls, period):
        year = int(period[2:6])
        week = int(period[7:])
        return iso_week_to_gregorian(year, week).strftime("Week of %b %d, %Y")


class QuarterlyPeriod(Period):

    prefix = "Q"
    validation_regex = r"\d{4}-(\d{1})$"
    minimum = 1
    maximum = 4
    contains = ["M", "W"]

    @classmethod
    def from_parts(cls, year, quarter):
        return "{}-{:d}-{:d}".format(cls.prefix, int(year), int(quarter))

    @classmethod
    def for_date(cls, date):
        quarter = 1 + (date.month - 1) // 3
        return cls.from_parts(date.year, quarter)

    @classmethod
    def start_end(cls, period):
        year = int(period[2:6])
        quarter = int(period[7])
        month = [None, 1, 4, 7, 10][quarter]
        end_month = calendar.monthrange(year, month + 2)[1]
        start = datetime.date(year, month, 1)
        end = datetime.date(year, month + 2, end_month)
        return start, end

    @classmethod
    def range(cls, start, stop, inclusive=False):
        cls.validate(start)
        cls.validate(stop)
        year_start = int(start[2:6])
        quarter_start = int(start[7])
        year_stop = int(stop[2:6])
        quarter_stop = int(stop[7])
        year = year_start
        quarter = quarter_start
        op = operator.le if inclusive else operator.lt
        while op((year, quarter), (year_stop, quarter_stop)):
            yield "{}-{:d}-{:d}".format(cls.prefix, year, quarter)
            quarter += 1
            if quarter == 5:
                quarter = 1
                year += 1
        raise StopIteration()

    @classmethod
    def display(cls, period):
        year = int(period[2:6])
        quarter = int(period[7])
        return "{}Q{}".format(year, quarter)


class MonthlyPeriod(Period):

    prefix = "M"
    validation_regex = r"\d{4}-(\d{2})$"
    minimum = 1
    maximum = 12
    contains = ["W"]

    @classmethod
    def for_date(cls, date):
        return "{}-{:d}-{:02d}".format(cls.prefix, date.year, date.month)

    @classmethod
    def start_end(cls, period):
        year = int(period[2:6])
        month = int(period[7:])
        end_month = calendar.monthrange(year, month)[1]
        start = datetime.date(year, month, 1)
        end = datetime.date(year, month, end_month)
        return start, end

    @classmethod
    def range(cls, start, stop, inclusive=False):
        cls.validate(start)
        cls.validate(stop)
        year_start = int(start[2:6])
        month_start = int(start[7:])
        year_stop = int(stop[2:6])
        month_stop = int(stop[7:])
        year = year_start
        month = month_start
        op = operator.le if inclusive else operator.lt
        while op((year, month), (year_stop, month_stop)):
            yield "{}-{:d}-{:02d}".format(cls.prefix, year, month)
            month += 1
            if month == 13:
                month = 1
                year += 1
        raise StopIteration()

    @classmethod
    def display(cls, period):
        year = int(period[2:6])
        month = int(period[7:])
        return datetime.date(year, month, 1).strftime("%B %Y")


class YearlyPeriod(Period):

    prefix = "Y"
    validation_regex = r"\d{4}$"
    contains = ["Q", "M", "W"]

    @classmethod
    def for_date(cls, date):
        return "{}-{:d}".format(cls.prefix, date.year)

    @classmethod
    def start_end(cls, period):
        year = int(period[2:])
        start = datetime.date(year, 1, 1)
        end = datetime.date(year, 12, 31)
        return start, end

    @classmethod
    def range(cls, start, stop, inclusive=False):
        cls.validate(start)
        cls.validate(stop)
        year_start = int(start[2:])
        year_stop = int(stop[2:])
        year = year_start
        op = operator.le if inclusive else operator.lt
        while op(year, year_stop):
            yield "{}-{:d}".format(cls.prefix, year)
            year += 1
        raise StopIteration()

    @classmethod
    def display(cls, period):
        year = int(period[2:])
        return "{}".format(year)


PREFIXES = {
    "W": WeeklyPeriod,
    "M": MonthlyPeriod,
    "Q": QuarterlyPeriod,
    "Y": YearlyPeriod
}

PERIOD_TYPES = {
    "weekly": WeeklyPeriod,
    "quarterly": QuarterlyPeriod,
    "monthly": MonthlyPeriod,
    "yearly": YearlyPeriod
}


PERIOD_PREFIXES = {
    period_type_class.prefix: period_type_class
    for period_type_class in PERIOD_TYPES.values()
}


def parse(value):
    """
    Convert:

    * 2015-W3, 2015-W03, 2015W03, 2015W3 ==> W-2015-03
    * 1/2015, 01/2015, Jan 2015, January 2015 ==> M-2015-01
    * 2015Q1 ==> Q-2015-1
    * 2015 ==> Y-2015
    """
    result = None
    if len(value) in [6, 7, 8] and value[:4].isdigit() and "W" in value and value.index("W") in [4, 5] and value[value.index("W") + 1:].isdigit():
        year, week = value[:4], value[value.index("W") + 1:]
        result = WeeklyPeriod.from_parts(year, week)
    elif len(value) == 6 and value[:4].isdigit() and value[4].lower() == "q" and value[5].isdigit():
        year, quarter = value[:4], value[5]
        result = QuarterlyPeriod.from_parts(year, quarter)
    elif len(value) == 4 and value.isdigit():
        result = YearlyPeriod.for_date(dateutil_parse(value))
    else:
        try:
            result = MonthlyPeriod.for_date(dateutil_parse(value))
        except ValueError:
            pass
    if result:
        validate(result)
    return result


def validate(raw_value):
    if raw_value[0] not in PERIOD_PREFIXES:
        raise ValidationError("invalid prefix in {}".format(raw_value))
    return PERIOD_PREFIXES[raw_value[0]].validate(raw_value)


def get_period(raw_value):
    if not raw_value:
        return
    if raw_value[0] not in PERIOD_PREFIXES:
        raise ValidationError("invalid prefix in {}".format(raw_value))
    return PERIOD_PREFIXES[raw_value[0]](raw_value)


def period_for_date(period_type, date=None):
    """
    for the given period_type, returns the period of the given date (or today
    if no second arg)
    """
    if date is None:
        date = datetime.datetime.now().date()
    return PERIOD_TYPES[period_type].for_date(date)


def period_start_end(period):
    """
    for the given period, return a tuple of the start and end dates
    """
    return PERIOD_PREFIXES[period[0]].start_end(period)


def period_range(start, stop, inclusive=False):
    """
    yields the periods from start to (but not including) stop.

    For example, period_range("Y-2010", "Y-2013") will yield "Y-2010",
    "Y-2011", "Y-2012".
    """
    if start[0] != stop[0]:
        raise ValidationError("start and stop must be of same period type")
    return PERIOD_PREFIXES[start[0]].range(start, stop, inclusive)


def period_display(period):
    """
    display the given period in a human-readable form
    """
    return PERIOD_PREFIXES[period[0]].display(period)


def iso_week_to_gregorian(iso_year, iso_week):
    fourth_jan = datetime.date(iso_year, 1, 4)
    year_start = fourth_jan - datetime.timedelta(fourth_jan.isoweekday() - 1)
    return year_start + datetime.timedelta(weeks=iso_week - 1)
